package net.knarcraft.bookswithoutborders;

import java.io.File;
import java.util.List;
import java.util.Objects;

import net.knarcraft.bookswithoutborders.state.EncryptionStyle;
import net.knarcraft.bookswithoutborders.utility.EncryptionHelper;
import net.knarcraft.bookswithoutborders.utility.FileHelper;
import org.bukkit.ChatColor;
import org.bukkit.Material;
import org.bukkit.Tag;
import org.bukkit.block.Sign;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.Action;
import org.bukkit.event.block.SignChangeEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.bukkit.event.player.PlayerItemHeldEvent;
import org.bukkit.event.player.PlayerJoinEvent;
import org.bukkit.inventory.EquipmentSlot;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.inventory.meta.BookMeta;
import org.bukkit.inventory.meta.ItemMeta;

/**
 * A listener for relevant events
 */
public class BooksWithoutBordersListener implements Listener {

    private final String slash = BooksWithoutBorders.SLASH;
    private final String bookFolderPath = BooksWithoutBorders.bwb.getDataFolder().getAbsolutePath() + slash + "Books";
    private final String bookFolder = bookFolderPath + slash;

    /**
     * Updates old books to a newer format
     *
     * @param player  <p>The player holding the book</p>
     * @param oldBook <p>Metadata about the held book</p>
     * @return <p>An updated book</p>
     */
    public ItemStack updateBook(Player player, BookMeta oldBook) {
        //handles hacked title-less books
        if (oldBook.getTitle() == null || oldBook.getTitle().length() < 3) {
            return null;
        }

        if (oldBook.getTitle().substring(oldBook.getTitle().length() - 3).equalsIgnoreCase("[U]")) {
            String fileName;

            if (oldBook.getAuthor() != null && oldBook.getAuthor().equalsIgnoreCase("unknown")) {
                //Unknown author is ignored
                fileName = oldBook.getTitle();
            } else {
                fileName = oldBook.getTitle() + BooksWithoutBorders.titleAuthorSeparator + oldBook.getAuthor();
            }

            String cleanPlayerName = BooksWithoutBorders.bwb.cleanString(player.getName());

            String[] possiblePaths = new String[]{
                    bookFolder + fileName + ".yml",
                    bookFolder + fileName + ".txt",
                    bookFolder + cleanPlayerName + slash + fileName + ".yml",
                    bookFolder + cleanPlayerName + slash + fileName + ".txt"
            };

            for (String path : possiblePaths) {
                File file = new File(path);
                if (file.isFile()) {
                    return BooksWithoutBorders.bwb.loadBook(player, fileName, "true", "player");
                }
            }
            return null;
        }

        return null;
    }

    @EventHandler
    public void onHold(PlayerItemHeldEvent event) {
        if (event.isCancelled()) {
            return;
        }

        Player player = event.getPlayer();
        int selectedSlot = event.getNewSlot();
        PlayerInventory playerInventory = player.getInventory();
        ItemStack selectedItem = playerInventory.getItem(selectedSlot);

        //Ignore irrelevant items
        if (selectedItem == null || selectedItem.getType() != Material.WRITTEN_BOOK) {
            return;
        }

        ItemMeta itemMetadata = selectedItem.getItemMeta();
        if (itemMetadata == null) {
            return;
        }

        //Update the book the user is viewing
        updateBookInHand(player, itemMetadata, true);
    }

    @EventHandler
    public void onPlayerJoin(PlayerJoinEvent event) {
        Player player = event.getPlayer();

        //Handle new players
        if (!BooksWithoutBorders.bwb.hasPlayedBefore(player.getName())) {
            boolean sendMessage = true;

            //Gives new players necessary books
            for (String bookName : BooksWithoutBorders.firstBooks) {
                sendMessage = giveBookToNewPlayer(bookName, player, sendMessage);
            }
        }

        //Updates any books in either hand
        ItemStack mainHandItem = player.getInventory().getItemInMainHand();
        ItemStack offHandItem = player.getInventory().getItemInOffHand();
        if (mainHandItem.getType() == Material.WRITTEN_BOOK) {
            ItemMeta itemMetadata = mainHandItem.getItemMeta();
            updateBookInHand(player, itemMetadata, true);
        }
        if (offHandItem.getType() == Material.WRITTEN_BOOK) {
            ItemMeta itemMetadata = offHandItem.getItemMeta();
            updateBookInHand(player, itemMetadata, false);
        }
    }

    @EventHandler
    public void onSignChange(SignChangeEvent event) {
        if (event.isCancelled()) {
            return;
        }

        String[] lines = event.getLines();
        Player player = event.getPlayer();

        //Check if creating a Books Without Borders Sign and if the player has permission
        if (!lines[0].equalsIgnoreCase("[BwB]") || !player.hasPermission("bookswithoutborders.signs")) {
            return;
        }

        //Mark the sign as active
        event.setLine(0, ChatColor.DARK_GREEN + "[BwB]");

        //Check if the sign is of a valid type
        if (!((lines[1].equalsIgnoreCase("[Encrypt]") || lines[1].equalsIgnoreCase("[Decrypt]") ||
                lines[1].equalsIgnoreCase("[Give]")) && lines[2].trim().length() > 0)) {
            //Mark the second line as invalid
            event.setLine(1, ChatColor.DARK_RED + lines[1]);
            return;
        }

        //Mark the second line as valid
        event.setLine(1, ChatColor.DARK_BLUE + lines[1]);

        //Mark valid encryption/decryption sign
        if (lines[1].equalsIgnoreCase(ChatColor.DARK_BLUE + "[Encrypt]") ||
                lines[1].equalsIgnoreCase(ChatColor.DARK_BLUE + "[Decrypt]")) {
            event.setLine(2, ChatColor.MAGIC + lines[2]);
            event.setLine(3, ChatColor.DARK_BLUE + lines[3]);
        }

        //Generate book giving sign
        if (lines[1].equalsIgnoreCase(ChatColor.DARK_BLUE + "[Give]")) {
            generateGiveSign(event, lines, player);
        }
    }

    @EventHandler
    public void onClick(PlayerInteractEvent event) {
        if (event.getClickedBlock() == null) {
            return;
        }

        Material clickedBlockType = event.getClickedBlock().getType();
        Player player = event.getPlayer();
        PlayerInventory playerInventory = player.getInventory();
        EquipmentSlot hand = event.getHand();
        if (hand == null) {
            return;
        }
        ItemStack heldItem = playerInventory.getItem(hand);
        Material heldItemType = heldItem.getType();

        if (event.getAction() == Action.RIGHT_CLICK_BLOCK && (Tag.SIGNS.isTagged(clickedBlockType) ||
                Tag.WALL_SIGNS.isTagged(clickedBlockType))) {
            //The player right-clicked a sign
            Sign sign = (Sign) event.getClickedBlock().getState();
            if (signLineEquals(sign, 0, "[BooksWithoutBorders]", ChatColor.DARK_GREEN)) {

                if (signLineEquals(sign, 1, "[Encrypt]", ChatColor.DARK_BLUE)) {
                    encryptHeldBookUsingSign(sign, heldItemType, player, hand);
                } else if (signLineEquals(sign, 1, "[Decrypt]", ChatColor.DARK_BLUE)) {
                    decryptHeldBookUsingSign(sign, heldItemType, player, hand);
                } else if (signLineEquals(sign, 1, "[Give]", ChatColor.DARK_BLUE) &&
                        getSignLineColor(sign, 2) == ChatColor.DARK_GREEN) {
                    giveBook(sign, player);
                }
            }
        } else if (heldItemType == Material.WRITTEN_BOOK && (event.getAction() == Action.LEFT_CLICK_AIR
                || event.getAction() == Action.LEFT_CLICK_BLOCK)) {
            BookMeta oldBook = (BookMeta) heldItem.getItemMeta();
            if (oldBook == null) {
                return;
            }

            decryptBook(oldBook, player, heldItem, hand);
        }
    }

    /**
     * Changes an edited sign into a sign to give books
     *
     * @param event  <p>The event caused by changing the sign</p>
     * @param lines  <p>The lines on the sign</p>
     * @param player <p>The player which edited the sign</p>
     */
    private void generateGiveSign(SignChangeEvent event, String[] lines, Player player) {
        if (lines[2].length() > 13 || lines[3].length() > 13) {
            player.sendMessage(ChatColor.RED + "[Give] signs' 3rd and 4th lines must be 13 characters or less!");
            markGiveSignValidity(event, false);
            return;
        }

        //Tests if a full file name has been supplied and points to an actual file
        String signFile = bookFolder + lines[2] + lines[3];
        if (FileHelper.bookFileExists(signFile)) {
            markGiveSignValidity(event, true);
        } else {
            if (isBookListIndex(lines[2])) {
                List<String> availableFiles = BooksWithoutBorders.bwb.listFiles(player, true, true);
                BooksWithoutBorders.loadList.put(player.getName(), availableFiles);
                markGiveSignValidity(event, true);
                return;
            }
        }
        markGiveSignValidity(event, false);
    }

    /**
     * Check whether an integer points to a public book
     *
     * @param possibleIndex <p>The string which might be a book index</p>
     * @return <p>True if the number is a book index</p>
     */
    private boolean isBookListIndex(String possibleIndex) {
        File bookDirectory = new File(bookFolderPath);

        try {
            //Tests if a load list number has been supplied
            int parsedLine = Integer.parseInt(possibleIndex);
            File[] foundFiles = bookDirectory.listFiles(File::isFile);

            return parsedLine >= 0 && parsedLine <= Objects.requireNonNull(foundFiles).length;
        } catch (NumberFormatException ignored) {
        }

        return false;
    }

    /**
     * Marks a give sign as valid or invalid
     *
     * @param event   <p>The event causing the creation of the give sign</p>
     * @param isValid <p>Whether the created sign is valid</p>
     */
    private void markGiveSignValidity(SignChangeEvent event, boolean isValid) {
        String[] lines = event.getLines();
        if (isValid) {
            event.setLine(2, ChatColor.DARK_GREEN + lines[2]);
            event.setLine(3, ChatColor.DARK_GREEN + lines[3]);
        } else {
            event.setLine(2, ChatColor.DARK_RED + lines[2]);
            event.setLine(3, ChatColor.DARK_RED + lines[3]);
        }
    }

    /**
     * Gives a book to a player joining for the first time
     *
     * @param bookName    <p>The name of the book to give</p>
     * @param player      <p>The player which just joined</p>
     * @param sendMessage <p>Whether to send a message to the joining player</p>
     * @return <p>True if a message has yet to be sent</p>
     */
    private boolean giveBookToNewPlayer(String bookName, Player player, boolean sendMessage) {
        if (!bookName.trim().isEmpty()) {

            //Handles loadList numbers
            if (isBookListIndex(bookName)) {
                List<String> availableFiles = BooksWithoutBorders.bwb.listFiles(player, true, true);
                BooksWithoutBorders.loadList.put(player.getName(), availableFiles);
            }

            //Give the book to the player if it exists
            ItemStack newBook = BooksWithoutBorders.bwb.loadBook(player, bookName, "true", "public");
            if (newBook != null) {
                player.getInventory().addItem(newBook);
            }

            //Send the player a welcome message if it exists
            if (!BooksWithoutBorders.welcomeMessage.trim().isEmpty() && newBook != null && sendMessage) {
                sendMessage = false;
                BooksWithoutBorders.bwb.getServer().getScheduler().scheduleSyncDelayedTask(BooksWithoutBorders.bwb,
                        () -> player.sendMessage(BooksWithoutBorders.welcomeMessage), 40L);
            }
        }
        return sendMessage;
    }

    /**
     * Updates a book in one of the player's hands
     *
     * @param player       <p>The player to update</p>
     * @param itemMetadata <p>Information about the held book</p>
     * @param mainHand     <p>Whether to update the book in the player's main hand</p>
     */
    private void updateBookInHand(Player player, ItemMeta itemMetadata, boolean mainHand) {
        PlayerInventory playerInventory = player.getInventory();
        ItemStack updatedBook = updateBook(player, (BookMeta) itemMetadata);
        if (updatedBook != null) {
            if (mainHand) {
                playerInventory.setItemInMainHand(updatedBook);
            } else {
                playerInventory.setItemInOffHand(updatedBook);
            }
        }
    }

    /**
     * Decrypts a book decryptable for a given group
     *
     * @param oldBook  <p>Metadata for the encrypted book</p>
     * @param player   <p>The player decrypting the book</p>
     * @param heldItem <p>The type of the held book</p>
     * @param hand     <p>The hand the player is using to hold the book</p>
     */
    private void decryptBook(BookMeta oldBook, Player player, ItemStack heldItem, EquipmentSlot hand) {
        ItemStack newBook;

        //Check if the book is encrypted by Books Without Borders
        if (!oldBook.hasLore() || oldBook.getLore() == null || !oldBook.getLore().get(0).contains(" encrypted]")) {
            return;
        }

        String groupName = oldBook.getLore().get(0).substring(3).split(" encrypted")[0];

        //Permission check
        if (!player.hasPermission("bookswithoutborders.decrypt." + groupName) &&
                !(BooksWithoutBorders.adminDecrypt && player.hasPermission("bookswithoutborders.admin"))) {
            return;
        }

        String fileName = oldBook.getTitle() + BooksWithoutBorders.titleAuthorSeparator + oldBook.getAuthor();

        String encryptionFile = BooksWithoutBorders.bwb.cleanString(groupName) + slash + fileName + ".yml";

        File file = new File(bookFolder + "Encrypted" + slash + encryptionFile);
        if (!file.isFile()) {
            file = new File(bookFolder + fileName + ".txt");
            if (!file.isFile()) {
                return;
            }
        }
        newBook = BooksWithoutBorders.bwb.loadBook(player, fileName, "true", groupName, heldItem.getAmount());

        player.getInventory().setItem(hand, newBook);
        player.closeInventory();
        player.sendMessage(ChatColor.GREEN + "Book auto-decrypted!");
    }

    /**
     * Encrypts and replaces the player's used book
     *
     * @param sign         <p>The clicked sign</p>
     * @param heldItemType <p>The item type used to click the sign</p>
     * @param player       <p>The player which clicked the sign</p>
     * @param hand         <p>The EquipmentSlot of the used hand</p>
     */
    private void encryptHeldBookUsingSign(Sign sign, Material heldItemType, Player player, EquipmentSlot hand) {
        ItemStack eBook;
        String[] lines = sign.getLines();
        boolean mainHand = hand == EquipmentSlot.HAND;
        if (heldItemType == Material.WRITTEN_BOOK) {
            player.closeInventory();
            eBook = BooksWithoutBorders.bwb.encryptBook(player, mainHand, ChatColor.stripColor(lines[2]),
                    EncryptionStyle.getFromString(ChatColor.stripColor(lines[3])));
            if (eBook != null) {
                player.getInventory().setItem(hand, eBook);
            }
        }
    }

    /**
     * Gives a player the book specified on a sign
     *
     * @param sign   <p>The sign the user clicked</p>
     * @param player <p>The player which clicked the sign</p>
     */
    private void giveBook(Sign sign, Player player) {
        String fileName = ChatColor.stripColor(sign.getLine(2));
        boolean isLoadListNumber = false;

        for (int x = 0; x < fileName.length(); x++) {
            if (!Character.isDigit(fileName.charAt(x)))
                break;
            if (x == fileName.length() - 1) {
                BooksWithoutBorders.loadList.put(player.getName(), BooksWithoutBorders.bwb.listFiles(player, true, true));
                isLoadListNumber = true;
            }
        }

        if (!isLoadListNumber && sign.getLine(3).length() >= 2)
            fileName = ChatColor.stripColor(sign.getLine(2)) + ChatColor.stripColor(sign.getLine(3));

        ItemStack newBook = BooksWithoutBorders.bwb.loadBook(player, fileName, "true", "public");

        if (newBook != null) {
            player.getInventory().addItem(newBook);
            player.sendMessage(ChatColor.GREEN + "Received book!");
        } else {
            player.sendMessage(ChatColor.RED + "Book failed to load!");
        }
    }

    /**
     * Decrypts and replaces the player's used book
     *
     * @param sign         <p>The clicked sign</p>
     * @param heldItemType <p>The item type used to click the sign</p>
     * @param player       <p>The player which clicked the sign</p>
     * @param hand         <p>The EquipmentSlot of the used hand</p>
     */
    private void decryptHeldBookUsingSign(Sign sign, Material heldItemType, Player player, EquipmentSlot hand) {
        //Decrypt the held book and replace it
        if (heldItemType == Material.WRITTEN_BOOK) {
            player.closeInventory();

            //Converts user supplied key into integer form
            String lineText = ChatColor.stripColor(sign.getLine(2));
            String key = EncryptionHelper.getNumberKeyFromStringKey(lineText);

            ItemStack book = BooksWithoutBorders.bwb.eLoad(player, key, false);
            if (book != null) {
                player.getInventory().setItem(hand, book);
                player.sendMessage(ChatColor.GREEN + "Book decrypted!");
            }
        }
    }

    /**
     * Gets the color of a line on a sign
     *
     * @param sign       <p>The sign to check</p>
     * @param lineNumber <p>The line number to check</p>
     * @return <p>The color of the sign</p>
     */
    private ChatColor getSignLineColor(Sign sign, int lineNumber) {
        return ChatColor.getByChar(sign.getLine(lineNumber).substring(0, 2));
    }

    /**
     * Checks if a line on a sign equals some string, and that the sign is the correct color
     *
     * @param sign       <p>The sign to read</p>
     * @param lineNumber <p>The sign line to read</p>
     * @param compareTo  <p>The string to look for</p>
     * @param color      <p>The color to match</p>
     * @return <p>True if the given string is what's on the sign</p>
     */
    private boolean signLineEquals(Sign sign, int lineNumber, String compareTo, ChatColor color) {
        String line = sign.getLine(lineNumber);
        return line.equalsIgnoreCase(color + compareTo);
    }

}
