package net.knarcraft.bookswithoutborders.utility;

import net.knarcraft.bookswithoutborders.state.BookHoldingState;
import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.state.HoldingItemState;
import net.knarcraft.bookswithoutborders.state.ItemSlot;
import org.bukkit.Material;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.BookMeta;

/**
 * The inventory helper mainly helps with getting and setting books
 */
public class InventoryHelper {

    /**
     * Gets the book the holder is playing
     *
     * @param player <p>The player holding the book</p>
     * @return <p>The book the player is holding</p>
     */
    public static ItemStack getHeldBook(Player player) {
        return getHeldBook(player, true);
    }

    /**
     * Gets the book the holder is playing
     *
     * @param player     <p>The player holding the book</p>
     * @param signedBook <p>Whether to check for signed or unsigned books</p>
     * @return <p>The book the player is holding</p>
     */
    public static ItemStack getHeldBook(Player player, boolean signedBook) {
        ItemSlot heldSlot = getHeldSlotBook(player, false, false, true, signedBook);
        if (heldSlot == ItemSlot.MAIN_HAND) {
            return player.getInventory().getItemInMainHand();
        } else if (heldSlot == ItemSlot.OFF_HAND) {
            return player.getInventory().getItemInOffHand();
        } else {
            throw new IllegalArgumentException("The player is not holding exactly one book.");
        }
    }

    /**
     * Replaces the book the player is holding
     *
     * @param player  <p>The player holding the book</p>
     * @param newBook <p>The new book the player should hold</p>
     */
    public static void setHeldWrittenBook(Player player, ItemStack newBook) {
        ItemSlot itemSlot = getHeldSlotBook(player, false, false, true, true);
        if (itemSlot == ItemSlot.MAIN_HAND) {
            replaceHeldItem(player, newBook, true);
        } else if (itemSlot == ItemSlot.OFF_HAND) {
            replaceHeldItem(player, newBook, false);
        } else {
            BooksWithoutBorders.sendErrorMessage(player, "Could not find a book to replace!");
        }
    }

    /**
     * Performs checks to validate that a player contains exactly one written
     *
     * @param player          <p>The player to validate</p>
     * @param noBookMessage   <p>The message to display if the player is not holding a book</p>
     * @param twoBooksMessage <p>The message to display if the player is holding one book in each hand</p>
     * @return <p>False if the player is holding exactly one written book</p>
     */
    public static boolean notHoldingOneWrittenBookCheck(Player player, String noBookMessage, String twoBooksMessage) {
        return notHoldingOneBookCheck(player, noBookMessage, twoBooksMessage, true);
    }

    /**
     * Performs checks to validate that a player contains exactly one writable
     *
     * @param player          <p>The player to validate</p>
     * @param noBookMessage   <p>The message to display if the player is not holding a book</p>
     * @param twoBooksMessage <p>The message to display if the player is holding one book in each hand</p>
     * @return <p>False if the player is holding exactly one writable book</p>
     */
    public static boolean notHoldingOneWritableBookCheck(Player player, String noBookMessage, String twoBooksMessage) {
        return notHoldingOneBookCheck(player, noBookMessage, twoBooksMessage, false);
    }

    /**
     * Performs checks to validate that a player contains exactly one book
     * @param player          <p>The player to validate</p>
     * @param noBookMessage   <p>The message to display if the player is not holding a book</p>
     * @param twoBooksMessage <p>The message to display if the player is holding one book in each hand</p>
     * @param written <p>Whether to require a written/signed book as opposed to a writable/unsigned book</p>
     * @return <p>False if the player is holding exactly one book</p>
     */
    private static boolean notHoldingOneBookCheck(Player player, String noBookMessage, String twoBooksMessage, boolean written) {
        BookHoldingState holdingState = getBookHoldingState(player);

        if (holdingState == BookHoldingState.NONE) {
            BooksWithoutBorders.sendErrorMessage(player, noBookMessage);
            return true;
        }

        if ((written && holdingState == BookHoldingState.SIGNED_BOTH_HANDS) ||
                (!written && holdingState == BookHoldingState.UNSIGNED_BOTH_HANDS)) {
            BooksWithoutBorders.sendErrorMessage(player, twoBooksMessage);
            return true;
        }

        return false;
    }

    /**
     * Check whether the player is holding one book
     *
     * @param player     <p>The player possibly holding a book</p>
     * @param signedBook <p>Whether to look for signed or unsigned books</p>
     * @return <p>True if the player is holding one book</p>
     */
    public static boolean isHoldingBook(Player player, boolean signedBook) {
        BookHoldingState holdingState = getBookHoldingState(player);

        if (holdingState == BookHoldingState.SIGNED_MAIN_HAND_UNSIGNED_OFF_HAND ||
                holdingState == BookHoldingState.UNSIGNED_MAIN_HAND_SIGNED_OFF_HAND) {
            return true;
        }
        if (signedBook) {
            return holdingState == BookHoldingState.SIGNED_MAIN_HAND ||
                    holdingState == BookHoldingState.SIGNED_OFF_HAND;
        } else {
            return holdingState == BookHoldingState.UNSIGNED_MAIN_HAND ||
                    holdingState == BookHoldingState.UNSIGNED_OFF_HAND;
        }
    }

    /**
     * Gets the slot of the player's held book
     * @param player <p>The player holding the book</p>
     * @param handMatters <p>Whether the differentiation between the main hand and the off hand is relevant</p>
     * @param mainHand <p>Whether to search the player's main hand or off hand, if it's relevant</p>
     * @param typeMatters <p>Whether the differentiation between signed and unsigned books is relevant</p>
     * @param writtenBook <p>Whether to search for written or unwritten books, if it's relevant</p>
     * @return <p>The slot of the player's held book</p>
     */
    public static ItemSlot getHeldSlotBook(Player player, boolean handMatters, boolean mainHand,
                                         boolean typeMatters, boolean writtenBook) {
        BookHoldingState state = getBookHoldingState(player);
        ItemStack mainHandItem = player.getInventory().getItemInMainHand();
        ItemStack offHandItem = player.getInventory().getItemInOffHand();
        Material requiredMaterial = writtenBook ? Material.WRITTEN_BOOK : Material.WRITABLE_BOOK;
        //Ambiguous or empty
        if (state == BookHoldingState.SIGNED_BOTH_HANDS ||
                state == BookHoldingState.UNSIGNED_BOTH_HANDS ||
                state == BookHoldingState.NONE) {
            return null;
        }
        if (handMatters && typeMatters) {
            if (mainHand && mainHandItem.getType() == requiredMaterial) {
                return ItemSlot.MAIN_HAND;
            } else if (!mainHand && offHandItem.getType() == requiredMaterial) {
                return ItemSlot.OFF_HAND;
            }
        } else if (handMatters) {
            //Ambiguous
            if (state == BookHoldingState.SIGNED_MAIN_HAND_UNSIGNED_OFF_HAND ||
                    state == BookHoldingState.UNSIGNED_MAIN_HAND_SIGNED_OFF_HAND) {
                return ItemSlot.NONE;
            }
            if (mainHand) {
                return ItemSlot.MAIN_HAND;
            } else {
                return ItemSlot.OFF_HAND;
            }
        } else if (typeMatters) {
            if ((writtenBook && (state == BookHoldingState.SIGNED_MAIN_HAND ||
                    state == BookHoldingState.SIGNED_MAIN_HAND_UNSIGNED_OFF_HAND)) || (!writtenBook &&
                    (state == BookHoldingState.UNSIGNED_MAIN_HAND ||
                            state == BookHoldingState.UNSIGNED_MAIN_HAND_SIGNED_OFF_HAND))) {
                return ItemSlot.MAIN_HAND;
            } else if ((writtenBook && (state == BookHoldingState.SIGNED_OFF_HAND ||
                    state == BookHoldingState.UNSIGNED_MAIN_HAND_SIGNED_OFF_HAND)) || (!writtenBook && (
                            state == BookHoldingState.UNSIGNED_OFF_HAND) ||
                    state == BookHoldingState.SIGNED_MAIN_HAND_UNSIGNED_OFF_HAND)) {
                return ItemSlot.OFF_HAND;
            }
        } else {
            if (state == BookHoldingState.SIGNED_MAIN_HAND || state == BookHoldingState.UNSIGNED_MAIN_HAND) {
                return ItemSlot.MAIN_HAND;
            } else if (state == BookHoldingState.SIGNED_OFF_HAND || state == BookHoldingState.UNSIGNED_OFF_HAND) {
                return ItemSlot.OFF_HAND;
            }
        }
        return ItemSlot.NONE;
    }

    /**
     * Gets the state of the player's book holding from all possible states
     * @param player <p>The player to check</p>
     * @return <p>The state of the player's book holding</p>
     */
    private static BookHoldingState getBookHoldingState(Player player) {
        ItemStack mainHandItem = player.getInventory().getItemInMainHand();
        ItemStack offHandItem = player.getInventory().getItemInOffHand();

        boolean hasSignedBookInMainHand = mainHandItem.getType() == Material.WRITTEN_BOOK;
        boolean hasUnsignedBookInMainHand = mainHandItem.getType() == Material.WRITABLE_BOOK;
        boolean hasSignedBookInOffHand = offHandItem.getType() == Material.WRITTEN_BOOK;
        boolean hasUnsignedBookInOffHand = offHandItem.getType() == Material.WRITABLE_BOOK;

        if (hasSignedBookInMainHand && hasSignedBookInOffHand) {
            return BookHoldingState.SIGNED_BOTH_HANDS;
        } else if (hasUnsignedBookInMainHand && hasUnsignedBookInOffHand) {
            return BookHoldingState.UNSIGNED_BOTH_HANDS;
        } else if (hasSignedBookInMainHand && hasUnsignedBookInOffHand) {
            return BookHoldingState.SIGNED_MAIN_HAND_UNSIGNED_OFF_HAND;
        } else if (hasUnsignedBookInMainHand && hasSignedBookInOffHand) {
            return BookHoldingState.UNSIGNED_MAIN_HAND_SIGNED_OFF_HAND;
        } else if (hasSignedBookInMainHand) {
            return BookHoldingState.SIGNED_MAIN_HAND;
        } else if (hasUnsignedBookInMainHand) {
            return BookHoldingState.UNSIGNED_MAIN_HAND;
        } else if (hasSignedBookInOffHand) {
            return BookHoldingState.SIGNED_OFF_HAND;
        } else if (hasUnsignedBookInOffHand) {
            return BookHoldingState.UNSIGNED_OFF_HAND;
        } else {
            return BookHoldingState.NONE;
        }
    }

    /**
     * Gets the state of which hands of the player contains an item type
     *
     * @param player <p>The player possibly holding the item</p>
     * @param type   <p>The type of material to look for</p>
     * @return <p>The state of the player's item holding</p>
     */
    private static HoldingItemState getHoldingItemState(Player player, Material type) {
        boolean hasBookInMainHand = player.getInventory().getItemInMainHand().getType() == type;
        boolean hasBookInOffHand = player.getInventory().getItemInOffHand().getType() == type;

        if (hasBookInMainHand && hasBookInOffHand) {
            return HoldingItemState.BOTH_HANDS;
        } else if (hasBookInMainHand) {
            return HoldingItemState.MAIN_HAND;
        } else if (hasBookInOffHand) {
            return HoldingItemState.OFF_HAND;
        } else {
            return HoldingItemState.NONE;
        }
    }

    /**
     * Gets metadata about the player's held book
     *
     * @param player   <p>The player holding the book</p>
     * @param mainHand <p>Whether to get information about a book in the player's main hand or off hand</p>
     * @return <p>Information about the held book</p>
     */
    public static BookMeta getHeldBookMetadata(Player player, boolean mainHand) {
        return (BookMeta) getHeldItem(player, mainHand).getItemMeta();
    }

    /**
     * Gets the item the player is holding
     *
     * @param player   <p>The player to get from</p>
     * @param mainHand <p>Whether to get the item in the player's main hand or off hand</p>
     * @return <p>The item the player is holding in the given hand</p>
     */
    public static ItemStack getHeldItem(Player player, boolean mainHand) {
        if (mainHand) {
            return player.getInventory().getItemInMainHand();
        } else {
            return player.getInventory().getItemInOffHand();
        }
    }

    /**
     * Replaces the player's held item
     *
     * @param player   <p>The player to replace the item for</p>
     * @param newBook  <p>The new book the player should hold</p>
     * @param mainHand <p>Whether to replace the item in the player's main hand or off hand</p>
     */
    public static void replaceHeldItem(Player player, ItemStack newBook, boolean mainHand) {
        if (mainHand) {
            player.getInventory().setItemInMainHand(newBook);
        } else {
            player.getInventory().setItemInOffHand(newBook);
        }
    }

}
